#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.api
# ---------------------------------------------------------------------
# Copyright (c) 2015 Merchise Autrement and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2015-10-19

'''Support the abstraction layer intended for API definitions.

There are two base classes for API definition:

- `Handler`:class:`\ : A base class -or ABC register- that allow to define
  interfaces and tools (in foundation and common layers) representing specific
  driver implementations.

- `Error`:class:`\ : Generic error bases that adopt different shapes in
  specific handler driver implementations.

WTF: Unfortunately, exception management don't work with ABCs in Python 3.
(See issue #\ 12029_).

A mechanism compatible with all versions for this pattern could be found in
this module.

.. _12029: http://bugs.python.org/issue12029

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


from xoutil.eight.meta import metaclass
from xoutil.eight.abc import ABC, ABCMeta


class Handler(ABC):
    '''Base class for all Xoutil Handlers.

    Handlers are concepts that require an API definition in low level layers
    (protocols), and specific implementations within concrete applications.

    The following convention will be used:

    - API Interface: Any of the `Handler` sub-classes that are declared inner
      foundation and common layers (for example, `Xoutil`).

    - API Tools: All non-empty methods and module -global- functions related
      to an API Interface.

    - API Kind: Any interface defining a core concept.  Normally are direct
      `Handler`:class: sub-classes and always have sub-classes in a concrete
      execution context.

    - Driver: `Handler` sub-classes that don't have further sub-classes and
      are unique in its kind within an execution context.

    - Fixing: Localization of the implementation for the driver given the
      handler kind.  The class-method `locate_driver_class`:meth: will obtain
      the proper driver class in the current execution context.

      Applications implementing drivers must use a mechanism of "Convention
      over Configuration" (CoC) to register their classes.  The simplest one
      is to guarantee that all modules containing needed drivers are loaded
      starting each execution context.

    - A driver could define a mechanism to validate if current execution
      context is applicable to its function or not.  See
      `is_applicable_for_context`:meth:.  This will support the fixing
      process.

      If more than one driver class is applicable for a handler kind within a
      context (even when someone -myself included- could say: "WTF"), an
      additional mechanism could be applied using the
      `get_importance_for_context`:meth: method.

    - Decorator `abstractmethod` use is an obligation indicator for methods
      must be implemented in the concrete layers.

    - Drivers could implement a set of "distinguishing features of its nature"
      (traits) that must be documented by redefining the
      `list_of_traits`:meth: method.  Each trait item will be a dictionary
      compliant value using the key names 'name' and 'type' as reserved and
      mandatory; and 'desc', and 'tags' as reserved and optional.

    '''
    def __str__(self):
        '''The "informal" string representation of a handler.'''
        return self._get_typed_desc()

    def __repr__(self):
        '''The "official" string representation of a handler.'''
        return '<{}>'.format(str(self))

    @classmethod
    def locate_driver_class(cls):
        '''Find a driver class for the given handler kind (cls).

        :param cls: Class representing a Handler Kind.

        This method must not be called in a driver.

        '''
        if cls is Handler:
            aux = list(get_leaf_classes(cls) - {cls})
            ok = is_driver_applicable_for_context
            drivers = [drv for drv in aux if ok(drv)]
            count = len(drivers)
            if count == 1:
                return drivers[0]
            elif count > 1:
                res, rank = None, None
                count = 0
                get_rank = get_driver_importance_for_context
                for drv in drivers:
                    new_rank = get_rank(drv)
                    if not res or rank < new_rank:
                        res, rank = drv, new_rank
                        count = 1
                    elif rank == new_rank:
                        count += 1
                if count == 1:
                    return res
                else:
                    assert count > 1
                    msg = 'Too many "{}" handler drivers obtained ({}).'
                    raise RuntimeError(msg.format(cls.__name__, count))
            else:
                msg = ('No "{}" handler driver found. Maybe you need to load '
                       'some missing modules.')
                raise RuntimeError(msg.format(cls.__name__))
        else:
            raise NotImplementedError('Method `locate_driver_class` is not '
                                      'implemented at `Handler` level.')

    @classmethod
    def is_applicable_for_context(cls):
        '''True if a driver is valid in current execution context.

        Must be redefined in drivers that could result invalids in some
        contexts.

        '''
        return True

    @classmethod
    def get_importance_for_context(cls):
        '''Rank of a driver.

        Must be redefined in drivers wishing to increase -or decrease- the
        level of importance.  The standard value is the integer "0".

        '''
        return 0

    @classmethod
    def list_of_traits(cls):
        '''Distinguishing features of specific handler driver nature.

        This method returns an iterator of dictionaries.  There are some
        reserved keys:

        - name: The feature name.

        - type: The kind or category of the feature.

        - desc: The feature description or documentation.

        - tags: Classifiers or descriptors for the feature.

        "name" and "type" are required.

        This is pure literate programming; if a feature is of type 'method'
        (or 'verb', a synonym) always you can call directly Python object
        methods representing these traits.

        '''
        return ()

    def _get_typed_desc(self):
        return '{}: {}'.format(type(self).__name__, self._get_untyped_desc())

    def _get_untyped_desc(self):
        from xoutil.object import multi_getter
        name, domain = multi_getter(self, 'name', 'domain')
        if name:
            return '{} ({})'.format(name, domain) if domain else name
        elif domain:
            return domain
        else:
            return hex(id(self))


class Error(Exception, ABC):
    '''Base class for all Xoutil API Exceptions.

    Allow to register concrete driver error classes in generic Xoutil
    exceptions.

    This class is an ABC in order to allow drivers to map concrete error
    classes using `adopt`:meth: method.

    For example::

      # --- In a general layer module (myapi) ---

      >>> from xoutil import api
      >>> @api.error
      ... class ResourceNotFoundError(Exception):
      ...    pass

      # --- In a driver implementation ---

      >>> import couchdb
      >>> myapi.ResourceNotFoundError.adopt(couchdb.http.ResourceNotFound)

      # --- In any code managing this exception (drivers) ---

      try:
          ...
      except api.error.ResourceNotFoundError:
          # catch registered exceptions (e.g. couchdb.http.ResourceNotFound)
          pass

    '''
    pass


def is_error_class(value):
    return isinstance(value, type) and issubclass(value, BaseException)


def _is_abc_nice_with_errors():
    '''See if ABCs works nice or not in this Python version.'''
    @Error.adopt
    class TestError(Exception):
        pass
    try:
        raise TestError('Any msg.')
    except Error:
        return True
    except Exception:
        return False

_ABC_NICE_WITH_ERRORS = _is_abc_nice_with_errors()


class _bag_assign(object):
    '''An execution context to block direct `ErrorBag` error assignments.'''
    @classmethod
    def enter(cls):
        '''Enter the context.

        Use as ``with _bag_assign.enter()``: ...

        '''
        from xoutil.context import context
        return context(cls)

    @classmethod
    def check(cls):
        '''Check if inner the context.'''
        from xoutil.context import context
        return bool(context[cls])


class ErrorBagMeta(ABCMeta):
    '''Meta-class for a bag of errors that need adaption.'''
    def __new__(cls, name, bases, attrs):
        if name == 'ErrorBag' and attrs.get('__module__') == __name__:
            _super = super(ErrorBagMeta, cls).__new__
            with _bag_assign.enter():
                return _super(cls, name, bases, attrs)
        else:
            raise RuntimeError('Only one instance of errors bag is allowed.')

    def __setattr__(self, name, value):
        is_err = is_error_class(value)
        if _bag_assign.check() or not (is_err or isinstance(value, self)):
            super(ErrorBagMeta, self).__setattr__(name, value)
        else:
            if is_err:
                kind = 'layer error'
                aux = value
            else:
                kind = 'internal error descriptor'
                aux = value.wrapped
            msg = 'API {} "{}" can\'t be directly assigned.'
            raise ValueError(msg.format(kind, aux.__name__))


class ErrorBag(metaclass(ErrorBagMeta)):
    '''Singleton for API error definitions (the bag).

    An API Error is something declared at a generic level and is mapped to
    concrete errors in specific implementations.  See `Error`:class: class for
    more information.

    Don't use this class directly with its name, use instead the declared
    alias "error".

    To adopt a generic error base, just use this class as a decorator; for
    example::

      >>> from xoutil import api
      >>> @api.error
      ... class ResourceNotFoundError(Exception):
      ...     pass

    '''
    def __new__(cls, wrapped):
        assert cls is ErrorBag
        if is_error_class(wrapped):
            if _ABC_NICE_WITH_ERRORS:
                self = wrapped
            else:
                self = super(ErrorBag, cls).__new__(cls)
                self.wrapped = wrapped
            name = wrapped.__name__
            if not hasattr(cls, name):
                with _bag_assign.enter():
                    # to avoid direct assigning block
                    setattr(cls, name, self)
                return wrapped
            else:
                raise RuntimeError('Error "{}" already adopted.'.format(name))
        else:
            from xoutil.names import simple_name
            msg = 'Parameter `wrapped` must be an exception class, not "{}".'
            raise TypeError(msg.format(simple_name(wrapped)))

    def __init__(self, wrapped):
        pass

    def __get__(self, obj, owner):
        import sys
        if obj is None:
            assert owner is ErrorBag
            cls = sys.exc_info()[0]
            if cls is not None and issubclass(cls, self.wrapped):
                return cls
            else:
                return self.wrapped


error = ErrorBag


def _abc_registered(cls):
    '''Internal class to get all ABC registered classes in a Python set.'''
    from abc import ABCMeta
    if isinstance(cls, ABCMeta):
        return set(cls._abc_registry)
    else:
        return set()


def get_leaf_classes(cls):
    '''List all branch sub-classes"

    A branch class don't have further sub-classes (either declared as Python
    inheritance or ABC registered).

    '''
    inner = set(type.__subclasses__(cls)) | _abc_registered(cls)
    if inner:
        res = set()
        for i in inner:
            if i not in res:
                for j in get_leaf_classes(i):
                    res.add(j)
        return res
    else:
        return {cls}


def is_driver_applicable_for_context(driver):
    '''Utility function to call `is_applicable_for_context` in drivers.

    If the `driver` was adopted using ABC, it's probable that miss the method
    `is_applicable_for_context`; in that case this function returns True.

    '''
    try:
        return driver.is_applicable_for_context()
    except AttributeError:
        return True


def get_driver_importance_for_context(driver):
    '''Utility function to call `get_importance_for_context` in drivers.

    If the `driver` was adopted using ABC, it's probable that miss the method
    `get_importance_for_context`; in that case this function returns 0.

    '''
    try:
        return driver.get_importance_for_context()
    except AttributeError:
        return 0



# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.types
# ---------------------------------------------------------------------
# Copyright (c) 2015 Merchise and Contributors
# Copyright (c) 2013, 2014 Merchise Autrement and Contributors
# Copyright (c) 2010-2012 Medardo Rodríguez
# All rights reserved.
#
# Author: Medardo Rodriguez
# Contributors: see CONTRIBUTORS and HISTORY file
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.


'''
This modules mirrors all the functions (and, in general, objects) from the
standard library module ``types``; but it also includes several new types and
type-related functions.
'''


from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        unicode_literals as _py3_unicode,
                        absolute_import as _py3_abs_imports)

from xoutil.modules import copy_members as _copy_python_module_members

_pm = _copy_python_module_members()
GeneratorType = _pm.GeneratorType
FunctionType = _pm.FunctionType
ModuleType = _pm.ModuleType

del _pm, _copy_python_module_members


from xoutil import Unset as _unset
from collections import Mapping

# FIXME: [med] Reintroduce UnsetType or deprecate it here.
from xoutil.logical import Logical as UnsetType    # noqa

from xoutil.names import strlist as strs
__all__ = strs('mro_dict', 'UnsetType', 'MappingProxyType',
               'SlotWrapperType', 'is_iterable', 'is_collection',
               'is_string_like', 'is_scalar', 'is_staticmethod',
               'is_classmethod', 'is_instancemethod', 'is_slotwrapper',
               'is_module', 'Required', 'NoneType', 'new_class',
               'prepare_class')
del strs


from .eight.types import (MappingProxyType, MemberDescriptorType,    # noqa
                          NoneType, MappingProxyType, SimpleNamespace,
                          DynamicClassAttribute, new_class, prepare_class,
                          _calculate_meta)


#: The type of methods that are builtin in Python.
#:
#: This is roughly the type of the ``object.__getattribute__`` method.
WrapperDescriptorType = SlotWrapperType = type(object.__getattribute__)

from xoutil.eight.exceptions import StandardError, BaseException    # noqa

# TODO: deprecate next
ExceptionBase = BaseException

from re import compile as _regex_compile

RegexPattern = type(_regex_compile(''))

del _regex_compile


def type_coerce(obj):
    '''Ensure return a valid type from `obj`.'''
    from xoutil.eight import class_types as ctypes
    return obj if isinstance(obj, ctypes) else obj.__class__


class mro_dict(Mapping):
    '''Utility behaving like a read-only dict of `target` MRO attributes.

    For example::

      >>> class A(object):
      ...     x = 12
      ...     y = 34

      >>> class B(A):
      ...     y = 56
      ...     z = 78

      >>> d = mro_dict(B)

      >>> d['x']
      12

      >>> d['y']
      56

      >>> d['z']
      78

    '''
    # TODO: What is the application for this utility?
    __slots__ = ('_probes', '_keys')

    def __init__(self, target):
        from xoutil.inspect import _static_getmro
        type_ = type_coerce(target)
        target_mro = _static_getmro(type_)
        self._probes = tuple(c.__dict__ for c in target_mro)
        self._keys = set()

    def __getitem__(self, name):
        from xoutil.objects import get_first_of
        result = get_first_of(self._probes, name, default=_unset)
        if result is not _unset:
            return result
        else:
            raise KeyError(name)

    def __iter__(self):
        if not self._keys:
            self._settle_keys()
        return iter(self._keys)

    def __len__(self):
        if not self._keys:
            self._settle_keys()
        return len(self._keys)

    def _settle_keys(self):
        for probe in self._probes:
            for key in probe:
                if key not in self._keys:
                    self._keys.add(key)


def mro_get_value_list(cls, name):
    '''Return a list with all `cls` class attributes in MRO.'''
    from xoutil.inspect import _static_getmro
    mro = _static_getmro(type_coerce(cls))
    return [t.__dict__[name] for t in mro if name in t.__dict__]


def mro_get_full_mapping(cls, name):
    '''Return a dictionary with all items from `cls` in MRO.

    All values corresponding to `name` must be valid mappings.

    '''
    aux = mro_get_value_list(cls, name)
    count = len(aux)
    if count == 0:
        return {}
    elif count == 1:
        return aux[0]
    else:
        res = {}
        for m in aux:
            for key in m:
                if key not in res:
                    res[key] = m[key]
        return res


# TODO: Many of is_*method methods here are needed to be compared against the
# standard lib's module inspect versions.  If they behave the same, these
# should be deprecated in favor of the standards.

def is_iterable(maybe):
    '''Returns True if `maybe` is an iterable object (e.g. implements the
    `__iter__` method):

    ::

        >>> is_iterable('all strings are iterable')
        True

        # Numbers are not
        >>> is_iterable(1)
        False

        >>> from xoutil.eight import range
        >>> is_iterable(range(1))
        True

        >>> is_iterable({})
        True

        >>> is_iterable(tuple())
        True

        >>> is_iterable(set())
        True

    '''
    try:
        iter(maybe)
    except:
        return False
    else:
        return True


def is_collection(maybe):
    '''Test `maybe` to see if it is a tuple, a list, a set or a generator
    function.

    It returns False for dictionaries and strings::

        >>> is_collection('all strings are iterable')
        False

        # Numbers are not
        >>> is_collection(1)
        False

        >>> from xoutil.eight import range
        >>> is_collection(range(1))
        True

        >>> is_collection({})
        False

        >>> is_collection(tuple())
        True

        >>> is_collection(set())
        True

        >>> is_collection(a for a in range(100))
        True

    .. versionchanged:: 1.5.5 UserList are collections.

    '''
    from xoutil.collections import UserList
    from xoutil.eight import range
    return isinstance(maybe, (tuple, range, list, set, frozenset,
                              GeneratorType, UserList))


def is_mapping(maybe):
    '''Test `maybe` to see if it is a valid mapping.'''
    return isinstance(maybe, (Mapping, MappingProxyType))


def is_string_like(maybe):
    '''Returns True if `maybe` acts like a string'''
    try:
        maybe + ""
    except TypeError:
        return False
    else:
        return True


def is_scalar(maybe):
    '''Returns True if `maybe` is a string, an int, or some other scalar type
    (i.e not an iterable.)

    '''
    return is_string_like(maybe) or not is_iterable(maybe)


def is_staticmethod(desc, name=_unset):
    '''Returns true if a `method` is a static method.

    This function takes the same arguments as :func:`is_classmethod`.

    '''
    if name:
        desc = mro_dict(desc).get(name, None)
    return isinstance(desc, staticmethod)


def is_classmethod(desc, name=_unset):
    '''Returns true if a `method` is a class method.

    :param desc: This may be the method descriptor or the class that holds the
           method, in the second case you must provide the `name` of the
           method.

           .. note::

              Notice that in the first case what is needed is the **method
              descriptor**, i.e, taken from the class' `__dict__`
              attribute. If instead you pass something like
              ``cls.methodname``, this method will return False whilst
              :func:`is_instancemethod` will return True.

    :param name: The name of the method, if the first argument is the class.

    '''
    if name:
        desc = mro_dict(desc).get(name, None)
    return isinstance(desc, classmethod)


def is_instancemethod(desc, name=_unset):
    '''Returns true if a given `method` is neither a static method nor a class
    method.

    This function takes the same arguments as :func:`is_classmethod`.

    '''
    if name:
        desc = mro_dict(desc).get(name, None)
    return isinstance(desc, FunctionType)


def is_slotwrapper(desc, name=_unset):
    '''Returns True if a given `method` is a slot wrapper (i.e. a method that
    is builtin in the `object` base class).

    This function takes the same arguments as :func:`is_classmethod`.

    '''
    if name:
        desc = mro_dict(desc).get(name, None)
    return isinstance(desc, SlotWrapperType)


def is_module(maybe):
    '''Returns True if `maybe` is a module.'''
    return isinstance(maybe, ModuleType)


class Required(object):
    '''A type for required fields in scenarios where a default is not
    possible.

    '''
    def __init__(self, *args, **kwargs):
        pass


# Real "Py4k" signature ``are_instances(*subjects, types)``.
def are_instances(*args):
    '''Return True if every `subject` is an instance of (any) `types`.

    :param subjects: All but last positional arguments.  Are the objects
        required to be instances of `types`.

    :param types: The last positional argument.  Either a single type or a
       sequence of types.  This must meet the conditions on the last argument
       of `isinstance`:func:.

    :returns: True or False.  True if for every `subject`,
       ``isinstance(subject, types)`` is True.  Otherwise, False.

    If no `subjects` are provided return True::

        >>> are_instances(int)
        True

    .. seealso:: The function :func:`no_instances` allows to test for subjects
                 not being instances of types.

    '''
    if not args:
        raise TypeError('are_instances requires at least an argument')
    subjects, types = args[:-1], args[-1]
    if not subjects:
        isinstance(None, types)   # HACK: always validate `types`.
    return all(isinstance(subject, types) for subject in subjects)


# Real Py4k signature ``are_instances(*subjects, types)``self.
def no_instances(*args):
    '''Return True if every `subject` is **not** an instance of (neither)
    `types`.

    :param subjects: All but last positional arguments.  Are the objects
        required not to be instances of `types`.

    :param types: The last positional argument.  Either a single type or a
       sequence of types.  This must meet the conditions on the last argument
       of `isinstance`:func:.

    :returns: True or False.  True if for every `subject`,
       ``isinstance(subject, types)`` is False.  Otherwise, False.

    If no `subjects` are provided return True::

        >>> no_instances(int)
        True

    .. note:: This is not the same as ``not are_instances(...)``.

       This function requires that *no* subject is an instance of `types`.
       Negating :func:`are_instances` would be True if *any* subject is not an
       instance of `types`.

    '''
    if not args:
        raise TypeError('no_instances requires at least an argument')
    subjects, types = args[:-1], args[-1]
    if not subjects:
        isinstance(None, types)   # HACK: always validate `types`.
    return all(not isinstance(subject, types) for subject in subjects)



# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.eight._types
# ---------------------------------------------------------------------
# Copyright (c) 2015 Merchise and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2015-06-10


'''Some functions implemented in module ``types`` in Python 3 but not in
Python 2 needed for '_meta*' implementation.

'''


from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_imports)


__all__ = [str(name) for name in ('new_class', 'prepare_class', )]


try:
    from types import new_class
except ImportError:
    # PEP 3115 compliant dynamic class creation.  Used in
    # xoutil.eight.meta.metaclass
    #
    # Taken from Python 3.3 code-base.
    def new_class(name, bases=(), kwds=None, exec_body=None):
        """Create a class object dynamically using the appropriate metaclass.

        """
        import sys
        meta, ns, kwds = prepare_class(name, bases, kwds)
        if exec_body is not None:
            exec_body(ns)
        if sys.version_info >= (3, 0):
            return meta(name, bases, ns, **kwds)
        else:
            return meta(name, bases, ns)


try:
    from types import prepare_class
except ImportError:
    def prepare_class(name, bases=(), kwds=None):
        """Call the __prepare__ method of the appropriate metaclass.

        Returns (metaclass, namespace, kwds) as a 3-tuple

        *metaclass* is the appropriate metaclass
        *namespace* is the prepared class namespace
        *kwds* is an updated copy of the passed in kwds argument with any
        'metaclass' entry removed. If no kwds argument is passed in, this will
        be an empty dict.

        """
        if kwds is None:
            kwds = {}
        else:
            kwds = dict(kwds)  # Don't alter the provided mapping
        meta = kwds.pop('metaclass', None)
        if not meta:
            if bases:
                meta = type(bases[0])
            else:
                meta = type
        if isinstance(meta, type):
            # when meta is a type, we first determine the most-derived
            # metaclass instead of invoking the initial candidate directly
            meta = _calculate_meta(meta, bases)
        if hasattr(meta, '__prepare__'):
            ns = meta.__prepare__(name, bases, **kwds)
        else:
            ns = {}
        return meta, ns, kwds


try:
    from types import _calculate_meta
except ImportError:
    def _calculate_meta(meta, bases):
        """Calculate the most derived metaclass."""
        winner = meta
        for base in bases:
            base_meta = type(base)
            if issubclass(winner, base_meta):
                continue
            if issubclass(base_meta, winner):
                winner = base_meta
                continue
            # else:
            raise TypeError("metaclass conflict: the metaclass of a derived "
                            "class must be a (non-strict) subclass of the "
                            "metaclasses of all its bases")
        return winner



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.uuid
# ---------------------------------------------------------------------
# Copyright (c) 2015, 2017 Merchise and Contributors
# Copyright (c) 2013, 2014 Merchise Autrement and Contributors
# Copyright (c) 2012 Medardo Rodríguez
# All rights reserved.
#
# Contributors: see CONTRIBUTORS and HISTORY file
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on Feb 17, 2012

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_imports)

from xoutil.names import strlist as strs
__all__ = strs('uuid', )
del strs

from xoutil.deprecation import deprecated
from xoutil.values.ids import str_uuid


@deprecated(str_uuid)
def uuid(random=False):
    '''Return a "Global Unique ID" as a string.

    :param random: If True, a random uuid is generated (does not use host id).

    '''
    return str_uuid(random)


del deprecated



# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.eight.types
# ---------------------------------------------------------------------
# Copyright (c) 2015 Merchise and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2015-02-25


'''All functions implemented in module ``types`` in Python 3 but not in
Python 2.

'''


from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        unicode_literals as _py3_unicode,
                        absolute_import as _py3_abs_imports)


__all__ = [str(name) for name in ('DictProxyType', 'MemberDescriptorType',
                                  'NoneType', 'MappingProxyType',
                                  'SimpleNamespace', 'DynamicClassAttribute',
                                  'new_class', 'prepare_class', )]


import sys
_py34 = sys.version_info >= (3, 4, 0)
del sys


from types import MemberDescriptorType, GetSetDescriptorType
if MemberDescriptorType is GetSetDescriptorType:    # As in pypy
    class _foo(object):
        __slots__ = 'bar'
    MemberDescriptorType = type(_foo.bar)
    del _foo


from ._types import new_class, prepare_class, _calculate_meta    # noqa


try:
    from types import NoneType
except ImportError:
    NoneType = type(None)


try:
    from types import DictProxyType
except ImportError:
    DictProxyType = type(object.__dict__)


try:
    from types import MappingProxyType
except ImportError:
    MappingProxyType = DictProxyType

# TODO: "MappingProxyType"" is named "DictProxyType" in Python 2.x.  Deprecate
# "DictProxyType" in favor of "MappingProxyType".


if _py34:
    from types import SimpleNamespace
else:
    # TODO: migrate to `xoutil.eight.reprlib`
    from abc import ABCMeta
    from .meta import metaclass

    from xoutil.reprlib import recursive_repr

    class SimpleNamespace(metaclass(ABCMeta)):
        '''A simple attribute-based namespace.

        SimpleNamespace(**kwargs)

        '''

        def __init__(self, **kwargs):
            self.__dict__.update(kwargs)

        def __eq__(self, other):
            # TODO: This method is not implemented in py33
            ok = isinstance(other, SimpleNamespace)
            return ok and self.__dict__ == other.__dict__

        @recursive_repr(str('namespace(...)'))
        def __repr__(self):
            keys = sorted(self.__dict__)
            items = ("{}={!r}".format(k, self.__dict__[k]) for k in keys)
            return "{}({})".format('namespace', ", ".join(items))

    del recursive_repr, ABCMeta, metaclass

    try:
        from types import SimpleNamespace as _sns
        SimpleNamespace.register(_sns)
        del _sns
    except ImportError:
        pass


try:
    from types import DynamicClassAttribute
except ImportError:
    class DynamicClassAttribute(object):
        """Route attribute access on a class to :meth:`~object.__getattr__`.

        This is a descriptor, used to define attributes that act differently
        when accessed through an instance and through a class.  Instance
        access remains normal, but access to an attribute through a class will
        be routed to the class's :meth:`~object.__getattr__` method; this is
        done by raising AttributeError.

        This allows one to have properties active on an instance, and have
        virtual attributes on the class with the same name (see
        :class:`~py3:enum.Enum` for an example).

        .. versionadded:: 1.5.5

        .. note:: The class Enum mentioned has not yet been backported.

        .. note:: In Python 3.4+ this is an alias to
                  :func:`types.DynamicClassAttribute
                  <py3:types.DynamicClassAttribute>`.

        """
        def __init__(self, fget=None, fset=None, fdel=None, doc=None):
            self.fget = fget
            self.fset = fset
            self.fdel = fdel
            # next two lines make DynamicClassAttribute act the same as
            # property
            self.__doc__ = doc or fget.__doc__
            self.overwrite_doc = doc is None
            # support for abstract methods
            _has_method = bool(getattr(fget, '__isabstractmethod__', False))
            self.__isabstractmethod__ = _has_method

        def __get__(self, instance, ownerclass=None):
            if instance is None:
                if self.__isabstractmethod__:
                    return self
                raise AttributeError()
            elif self.fget is None:
                raise AttributeError("unreadable attribute")
            return self.fget(instance)

        def __set__(self, instance, value):
            if self.fset is None:
                raise AttributeError("can't set attribute")
            self.fset(instance, value)

        def __delete__(self, instance):
            if self.fdel is None:
                raise AttributeError("can't delete attribute")
            self.fdel(instance)

        def getter(self, fget):
            fdoc = fget.__doc__ if self.overwrite_doc else None
            cls = type(self)
            result = cls(fget, self.fset, self.fdel, fdoc or self.__doc__)
            result.overwrite_doc = self.overwrite_doc
            return result

        def setter(self, fset):
            result = type(self)(self.fget, fset, self.fdel, self.__doc__)
            result.overwrite_doc = self.overwrite_doc
            return result

        def deleter(self, fdel):
            result = type(self)(self.fget, self.fset, fdel, self.__doc__)
            result.overwrite_doc = self.overwrite_doc
            return result



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.eight
# ---------------------------------------------------------------------
# Copyright (c) 2015 Merchise and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under
# the terms of the LICENCE attached in the distribution package.
#
# Created on 2015-02-26

'''Xoutil extensions for writing code that runs on Python 2 and 3

The name comes from (Manu's idea') "2 raised to the power of 3".

There is an existing library written by "Benjamin Peterson" named `six`_, both
(`xoutil.eight` and `six`) can be used together since this module don't claim
to be a replacement of `six`, just some extra extensions.  Nevertheless, there
are some simple definitions that even when are in `six` also are defined also
here.

This package also fixes some issues from PyPy interpreter.

.. _six: https://pypi.python.org/pypi/six

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        unicode_literals as _py3_unicode,
                        absolute_import)

import sys

# Python versions

_py2 = sys.version_info[0] == 2
_py3 = sys.version_info[0] == 3
_py33 = sys.version_info >= (3, 3, 0)
_pypy = sys.version.find('PyPy') >= 0

del sys

try:
    from hashlib import sha1 as sha
except ImportError:
    from sha import sha    # noqa

try:
    base_string = basestring
    string_types = (str, unicode)
except NameError:
    base_string = str
    string_types = (str, )

if _py3:
    integer_types = int,
    class_types = type,
    text_type = str
    unichr = chr
else:
    from types import ClassType
    integer_types = (int, long)
    class_types = (type, ClassType)
    text_type = unicode
    unichr = unichr
    del ClassType

binary_type = bytes


def typeof(obj):
    '''Obtain the object's type compatible with Py 2**3.'''
    if _py3:
        return type(obj)
    else:
        from types import InstanceType as OldClass
        return obj.__class__ if isinstance(obj, OldClass) else type(obj)


try:
    __intern = intern

    def intern(string):
        # Avoid problems in Python 2.x when using unicode by default.
        return __intern(str(str() + string))

    intern.__doc__ = __intern.__doc__
except NameError:
    from sys import intern    # noqa


if _py3:
    input = input
    range = range
    zip = zip
else:
    range = xrange
    input = raw_input
    from itertools import izip as zip    # noqa


def iterkeys(d):
    '''Return an iterator over the keys of a dictionary.'''
    return (d.keys if _py3 else d.iterkeys)()


def itervalues(d):
    '''Return an iterator over the values of a dictionary.'''
    return (d.values if _py3 else d.itervalues)()


def iteritems(d):
    '''Return an iterator over the (key, value) pairs of a dictionary.'''
    return (d.items if _py3 else d.iteritems)()


try:
    callable = callable
except NameError:
    def callable(obj):
        return any('__call__' in cls.__dict__ for cls in type(obj).__mro__)


if _py3:
    from io import StringIO
else:
    from StringIO import StringIO    # noqa


if _py3:
    import builtins
    exec_ = getattr(builtins, 'exec')  # noqa
else:
    def exec_(_code_, _globs_=None, _locs_=None):
        """Execute code in a namespace."""
        import sys
        if _globs_ is None:
            frame = sys._getframe(1)
            _globs_ = frame.f_globals
            if _locs_ is None:
                _locs_ = frame.f_locals
            del frame
        elif _locs_ is None:
            _locs_ = _globs_
        exec("""exec _code_ in _globs_, _locs_""")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.values
# ---------------------------------------------------------------------
# Copyright (c) 2015 Merchise and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2015-07-15

'''Some generic coercers (or checkers) for value types.

This module coercion function are not related in any way to deprecated old
python feature, are similar to a combination of object mold/check:

- *Mold* - Fit values to expected conventions.

- *Check* - These functions must return `Invalid`\ [#pyni]_ special value to
  specify that expected fit is not possible.

.. [#pyni] We don't use Python classic `NotImplemented` special value in order
           to obtain False if the value is not coerced (`Invalid`).

A custom coercer could be created with closures, for an example see
`create_int_range_coerce`:func:.

Also contains sub-modules to obtain, convert and check values of common types.

.. versionadded:: 1.7.0

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


import sys
from re import compile as regex_compile
from xoutil.eight.abc import ABCMeta
from xoutil.eight.meta import metaclass
from xoutil.functools import lwraps
from xoutil.logical import Logical as InvalidType


Invalid = InvalidType('Invalid')


_coercer_decorator = lwraps(__coercer__=True)


def valid(arg):
    '''True if `arg` is not Invalid.'''
    return arg is not Invalid


def check(coerce, arg):
    '''Coerce `arg`; if invalid raises a TypeError.

    See `compose`:class: and `some`:class: on how to combine "AND" and "OR"
    coercers.

    See `create_int_range_coerce`:func: for an example.

    '''
    res = coerce(arg)
    if valid(res):
        return res
    else:
        msg = 'Value "{}" is not coerced by "{}".'
        raise TypeError(msg.format(arg, coercer_name(arg)))


class MetaCoercer(ABCMeta):
    '''Meta-class for `coercer`:class:.

    This meta-class allows that several objects are considered valid instances
    of `coercer`:class:\ :

    - Functions decorated with `coercer`:class: (used with its decorator
      facet).

    - Instances of any sub-class of `custom`:class:.

    - Instances of `coercer`:class: itself.

    See the class declaration (`coercer`:class:) for more information.

    '''
    def __instancecheck__(self, instance):
        return (getattr(instance, '__coercer__', False) or
                super(MetaCoercer, self).__instancecheck__(instance))


class coercer(metaclass(MetaCoercer)):
    '''Special coercer class.

    This class has several facets:

    - Pure type-checkers when a type or tuple of types are received as
      argument.  See `istype`:class: for more information.

    - Return equivalent coercer from some special values:

      * None -> identity_coerce

      * () -> void_coerce

    - A decorator for functions; when a function is given, decorate it to
      become a coercer.  The mark itself is not enough, functions intended to
      be coercers must fulfills the protocol (not to produce exception and
      return `Invalid` on fails).  For example::

        >>> @coercer
        ... def age_coerce(arg):
        ...     res = int_coerce(arg)
        ...     return res if valid(res) and 0 < arg <= 120 else Invalid

        >>> isinstance(age_coerce, coercer)
        True

    '''
    __slots__ = ()

    def __new__(cls, source):
        from types import FunctionType as function
        if source is None:
            return identity_coerce
        elif source is ():
            return void_coerce
        elif isinstance(source, coercer):
            return source
        elif isinstance(source, (function, staticmethod, classmethod)):
            return _coercer_decorator(source)
        else:
            inner = types_tuple_coerce(source)
            return istype(inner) if inner else Invalid


def coercer_name(arg, join=None):
    '''Get the name of a coercer.

    :param arg: Coercer to get the name.  Also processes collections (tuple,
           list or set) of coercers.  Any other value is considered invalid
           and raises an exception.

    :param join: When a collection is used; if this argument is None a
           collection of names is returned, if not None then is used to join
           the items in a resulting string.  For example::

             >>> coercer_name((int_coerce, float_coerce))
             ('int', 'float')

             >>> coercer_name((int_coerce, float_coerce), '_')
             'int_float'

           To obtain pretty-print tuples, use something like::

             >>> coercer_name((int_coerce, float_coerce),
             ...              join=lambda arg: '(%s)' % ', '.join(arg))

    This function not only works with coercers, all objects that fulfill
    needed protocol to get names will also be valid.

    '''
    # TODO: Maybe this function must be moved to `xoutil.names`
    from xoutil.eight import string_types
    if isinstance(arg, (tuple, list, set)):
        res = type(arg)(coercer_name(c) for c in arg)
        if isinstance(join, string_types):
            join = join.join
        return str(join(res)) if join else res
    else:
        try:
            res = arg.__name__
        except BaseException:
            res = str(arg)
        suffix = str('_coerce')
        if res.endswith(suffix):
            res = res[:-len(suffix)]
        return res


@coercer
def identity_coerce(arg):
    'Leaves unchanged the passed argument `arg`.'
    return arg


@coercer
def void_coerce(arg):
    '''Always `Invalid`.'''
    return Invalid


@coercer
def type_coerce(arg):
    '''Check if `arg` is a valid type.'''
    from xoutil.eight import class_types
    return arg if isinstance(arg, class_types) else Invalid


@coercer
def types_tuple_coerce(arg):
    '''Check if `arg` is valid for `isinstance` or `issubclass` 2nd argument.

    Type checkers are any class, a type or tuple of types.  For example::

      >>> types_tuple_coerce(object) == (object,)
      True

      >>> types_tuple_coerce((int, float)) == (int, float)
      true

      >>> types_tuple_coerce('not-a-type') is Invalid
      True

    See `type_coerce` for more information.

    '''
    if valid(type_coerce(arg)):
        return (arg, )
    elif isinstance(arg, tuple) and all(valid(type_coerce(t)) for t in arg):
        return arg
    else:
        return Invalid


@coercer
def callable_coerce(arg):
    '''Check if `arg` is a callable object.'''
    from xoutil.eight import callable
    return arg if callable(arg) else Invalid


@coercer
def file_coerce(arg):
    '''Check if `arg` is a file-like object.'''
    from xoutil.eight.io import is_file_like
    return arg if is_file_like(arg) else Invalid


@coercer
def float_coerce(arg):
    '''Check if `arg` is a valid float.

    Other types are checked (string, int, complex).

    '''
    from xoutil.eight import integer_types, string_types
    if isinstance(arg, float):
        return arg
    elif isinstance(arg, integer_types):
        return float(arg)
    elif isinstance(arg, string_types):
        try:
            return float(arg)
        except ValueError:
            return Invalid
    elif isinstance(arg, complex):
        return arg.real if arg.imag == 0 else Invalid
    else:
        return Invalid


@coercer
def int_coerce(arg):
    '''Check if `arg` is a valid integer.

    Other types are checked (string, float, complex).

    '''
    from xoutil.eight import integer_types
    if isinstance(arg, integer_types):
        return arg
    else:
        arg = float_coerce(arg)
        if valid(arg):
            res = int(arg)
            return res if arg - res == 0 else Invalid
        else:
            return Invalid


@coercer
def number_coerce(arg):
    '''Check if `arg` is a valid number (integer or float).

    Types that are checked (string, int, float, complex).

    '''
    from xoutil.eight import integer_types
    if isinstance(arg, integer_types):
        return arg
    else:
        f = float_coerce(arg)
        if valid(f):
            i = int(f)
            return i if f - i == 0 else f
        else:
            return Invalid


@coercer
def positive_int_coerce(arg):
    '''Check if `arg` is a valid positive integer.'''
    res = int_coerce(arg)
    return res if res is Invalid or res >= 0 else Invalid


def create_int_range_coerce(min, max):
    '''Create a coercer to check integers between a range.'''
    min, max = check(int_coerce, min), check(int_coerce, max)
    if min < max:
        @coercer
        def inner(arg):
            'Check if `arg` is a valid integer between "{}" and "{}".'
            arg = int_coerce(arg)
            if valid(arg) and min <= arg <= max:
                return arg
            else:
                return Invalid
        inner.__name__ = str('int_between_{}_and_{}_coerce'.format(min, max))
        inner.__doc__ = inner.__doc__.format(min, max)
        return inner
    else:
        msg = '"{}" must be less than or equal "{}"'
        raise ValueError(msg.format(min, max))


# Identifiers and strings

# TODO: In Py3k "ña" is a valid identifier and this regex won't allow it
_IDENTIFIER_REGEX = regex_compile('(?i)^[_a-z][\w]*$')


@coercer
def identifier_coerce(arg):
    '''Check if `arg` is a valid Python identifier.

    .. note:: Only Python 2's version of valid identifier. This means that
              some Python 3 valid identifiers are not considered valid.  This
              helps to keep things working the same in Python 2 and 3.

    '''
    from xoutil.eight import string_types
    ok = isinstance(arg, string_types) and _IDENTIFIER_REGEX.match(arg)
    return str(arg) if ok else Invalid


_FULL_IDENTIFIER_REGEX = regex_compile('(?i)^[_a-z][\w]*([.][_a-z][\w]*)*$')


@coercer
def full_identifier_coerce(arg):
    '''Check if `arg` is a valid dotted Python identifier.

    See `identifier_coerce`:func: for what "validity" means.

    '''
    from xoutil.eight import string_types
    ok = isinstance(arg, string_types) and _FULL_IDENTIFIER_REGEX.match(arg)
    return str(arg) if ok else Invalid


@coercer
def names_coerce(arg):
    '''Check if `arg` is a valid tuple of valid object names (identifiers).

    If only one string is given, is returned as the only member of the
    resulting tuple.

    '''
    from xoutil.eight import string_types
    arg = (arg,) if isinstance(arg, string_types) else tuple(arg)
    return iterable(identifier_coerce)(arg)


# == Iterators ==

def create_unique_member_coerce(coerce, container):
    '''Useful to wrap member coercers when coercing containers.

    See `iterable`:class: and `mapping`:class:.

    Resulting coercer check that a member must be unique (not repeated) after
    it's coerced.

    For example::

      >>> from xoutil.values import (mapping,
      ...                            create_unique_member_coerce,
      ...                            int_coerce, float_coerce)

      >>> sample = {'1': 1, 2.0: '3', 1.0 + 0j: '4.1'}

      >>> dc = mapping(int_coerce, float_coerce)
      >>> dc(dict(sample))
      {1: 1.0, 2: 3.0}

      >>> dc = mapping(create_unique_member_coerce(int_coerce), float_coerce)
      >>> dc(dict(sample))
      Invalid

    '''
    coerce = check(coercer, coerce)

    @coercer
    def inner(arg):
        '''Check a member with "{}" coercer and warrant that is unique.'''
        # assert arg in container
        res = coerce(arg)
        if valid(res) and hash(res) != hash(arg) and res in container:
            res = Invalid
        return res

    cname = coercer_name(coerce)
    inner.__name__ = str('unique_member_{}_coerce'.format(cname))
    inner.__doc__ = inner.__doc__.format(cname)
    return inner


@coercer
def sized_coerce(arg):
    '''Return a valid sized iterable from `arg` (if valid).

    If `arg` is iterable but not sized, is converted to a list.  For example::

      >>> sized_coerce(i for i in range(1, 10, 2))
      [1, 3, 5, 7, 9]

      >>> s = {1, 2, 3}
      >>> sized_coerce(s) is s
      True

    '''
    from xoutil.collections import Iterable, Sized
    if isinstance(arg, Iterable):
        return arg if isinstance(arg, Sized) else list(arg)
    else:
        return Invalid


@coercer.adopt
class custom(object):
    '''Base class for any custom coercer.

    The field `inner` stores an internal data used for the custom coercer;
    could be a callable, an inner coercer, or a tuple of inner checkers if
    more than one is needed, ...

    The field `scope` stores the exit (not regular) condition: the value that
    fails or -if needed- a tuple with (exit-value, exit-coercer) or
    (error-value, error).  The exit condition is not always a failure, for
    example in `some`:class: it is the one that is valid among other inner
    coercers.  To understand better this think on (AND, OR) operators a chain
    of ANDs exits with the first failure and a chains of ORs exits with the
    first success.

    All custom coercers are callable (must redefine `__call__`:meth:)
    receiving one argument that must be coerced.  For example::

      >>> def foobar(*args):
      ...     coerce = pargs(int_coerce)
      ...     return coerce(args)

    This class has two protected fields (`_str_join` and `_repr_join`) that
    are used to call `coercer_name`:func: in `__str__`:meth: and
    `__repr__`:meth: special methods.

    '''
    __slots__ = ('inner', 'scope')

    _str_join = '_'
    _repr_join = ', '

    def __init__(self, *args, **kwargs):
        # This constructor is a placeholder for those custom coercers that can
        # return an instance of a different type in the `__new__`:meth:.
        from xoutil import Unset
        self.scope = Unset

    def __str__(self):
        name = coercer_name(self.inner, join=self._str_join)
        cls_name = type(self).__name__
        return str('{}_{}_coerce'.format(name, cls_name))

    def __repr__(self):
        name = coercer_name(self.inner, join=self._repr_join)
        cls_name = type(self).__name__
        return str('{}({})'.format(cls_name, name))

    def __call__(self, arg):
        return Invalid


class istype(custom):
    '''Pure type-checker.

    It's constructed from an argument valid for `types_tuple_coerce`:func:
    coercer.

    For example::

        >>> int_coerce = istype(int)

        >>> int_coerce(1)
        1

        >>> int_coerce('1')
        Invalid

        >>> number_coerce = istype((int, float, complex))

        >>> number_coerce(1.25)
        1.25

        >>> number_coerce('1.25')
        Invalid

    '''
    __slots__ = ('inner', 'scope')

    def __init__(self, types):
        super(istype, self).__init__()
        self.inner = check(types_tuple_coerce, types)

    def __call__(self, arg):
        return arg if isinstance(arg, self.inner) else Invalid


class typecast(istype):
    '''A type-caster.

    It's constructed from an argument valid for `types_tuple_coerce`:func:
    coercer.  Similar to `istype`:class: but try to convert the value if
    needed.

    For example::

        >>> int_cast = typecast(int)

        >>> int_cast('1')
        1

        >>> int_cast('1x')
        Invalid

    '''
    def __call__(self, arg):
        res = super(typecast, self).__call__(arg)
        i = 0
        while not valid(res) and i < len(self.inner):
            try:
                tp = self.inner[i]
                res = tp(arg)
                self.scope = tp
            except BaseException:
                i += 1
        return res


class safe(custom):
    '''Uses a function (or callable) in a safe way.

    Receives any callable that expects only one argument and returns another
    value; if that returned value is a Boolean, those functions are called
    predicates (See `xoutil.connote`:mod: for more information).

    The wrapped function is called in a safe way (inside try/except); if an
    exception is raised the coercer returns `Invalid` and it is saved in the
    instance field `scope`.

    '''
    def __init__(self, func):
        super(safe, self).__init__()
        self.inner = check(func, callable_coerce)

    def __call__(self, arg):
        try:
            return arg if self.inner(arg) else Invalid
        except BaseException as error:
            self.scope = (arg, error)
            return Invalid


class compose(custom):
    '''Returns the composition of several inner `coercers`.

    ``compose(f1, ... fn)`` is equivalent to f1(...(fn(arg))...)``.

    If no coercer is given return `identity_coerce`:func:.

    Could be considered an "AND" operator with some light differences because
    the nature of coercers: ordering the coercers is important when some can
    modify (adapt) original values.

    '''

    def __new__(cls, *coercers):
        coercers = pargs(coercer)(coercers)
        inner = tuple(check(coercer, c) for c in coercers
                      if c is not identity_coerce)
        if len(inner) > 1:
            self = super(compose, cls).__new__(cls)
            super(compose, self).__init__()
            self.inner = inner
            return self
        elif len(inner) == 1:
            return inner[0]
        else:
            return identity_coerce

    def __call__(self, arg):
        coercers = self.inner
        i = 0
        res = arg
        ok = True
        while ok and i < len(coercers):
            coerce = coercers[i]
            aux = coerce(res)
            if valid(aux):
                i += 1
            else:
                ok = False
                self.scope = (res, coerce)
            res = aux
        return res


class some(custom):
    '''Represent OR composition of several inner `coercers`.

    ``compose(f1, ... fn)`` is equivalent to f1(arg) or f2(arg) ... fn(arg)``
    in the sense "the first not `Invalid`".

    If no coercer is given return `void_coerce`:func:.

    '''

    def __new__(cls, *coercers):
        coercers = pargs(coercer)(coercers)
        inner = tuple(check(coercer, c) for c in coercers
                      if c is not void_coerce)
        if len(inner) > 1:
            self = super(some, cls).__new__(cls)
            super(some, self).__init__()
            self.inner = inner
            return self
        elif len(inner) == 1:
            return inner[0]
        else:
            return void_coerce

    def __init__(self, *coercers):
        pass

    def __call__(self, arg):
        coercers = self.inner
        i = 0
        res = Invalid
        while res is Invalid and i < len(coercers):
            coercer = coercers[i]
            value = coercer(arg)
            if valid(value):
                res = value
                self.scope = coercer
            else:
                i += 1
        return res


class combo(custom):
    '''Represent a zip composition of several inner `coercers`.

    An instance of this class is constructed from a sequence of coercers and
    the its purpose is coerce a sequence of values.  Return a sequence\
    [#type]_ where each item contains the i-th element from applying the i-th
    coercer to the i-th value from argument sequence::

      coercers -> (coercer-1, coercer-2, ... )
      values -> (value-1, value-2, ... )
      combo(coercers)(values) -> (coercer-1(value-1), coercer-2(value-2), ...)

    If any value is coerced invalid, the function returns `Invalid` and the
    combo's instance variable `scope` receives the duple ``(failed-value,
    failed-coercer)``.

    The returned sequence is truncated in length to the length of the shortest
    sequence (coercers or arguments).

    If no coercer is given, all sequences are coerced as empty.

    .. [#type] The returned sequence is of the same type as the argument
               sequence if possible.

    '''

    def __init__(self, *coercers):
        super(combo, self).__init__()
        coercers = pargs(coercer)(coercers)
        self.inner = tuple(check(coercer, c) for c in coercers)

    def __call__(self, arg):
        from xoutil import Unset
        from xoutil.collections import Iterable
        if isinstance(arg, Iterable):
            coercers = self.inner
            items = iter(arg)
            i = 0
            res = []
            ok = True
            while valid(res) and ok and i < len(coercers):
                item = next(items, Unset)
                if item is not Unset:
                    coerce = coercers[i]
                    value = coerce(item)
                    if valid(value):
                        res.append(value)
                        i += 1
                    else:
                        res = Invalid
                        self.scope = (item, coerce)
                else:
                    ok = False
            if valid(res):
                try:
                    res = type(arg)(res)
                except BaseException:
                    pass
        else:
            res = Invalid
        return res


class pargs(custom):
    '''Create a inner coercer that check variable argument passing.

    Created coercer closure must always receives an argument that is an valid
    iterable with all members coerced properly with the argument of this outer
    creator function.

    If the inner closure argument has only a member and this one is not
    properly coerced but it's an iterabled with all members that coerced well,
    this member will be the assumed iterable instead the original argument.

    In the following example::

      >>> from xoutil.values import (iterable, int_coerce)

      >>> def foobar(*args):
      ...     coerce = iterable(int_coerce)
      ...     return coerce(args)

      >>> args = (1, 2.0, '3.0')
      >>> foobar(*args)
      (1, 2, 3)

      >>> foobar(args)
      Invalid

    An example using `pargs`:class:\ ::

      >>> from xoutil.values import (pargs, int_coerce)

      >>> def foobar(*args):
      ...     # Below, "coercer" receives the returned "inner"
      ...     coerce = pargs(int_coerce)
      ...     return coerce(args)

      >>> args = (1, 2.0, '3.0')
      >>> foobar(*args)
      (1, 2, 3)

      >>> foobar(args)
      (1, 2, 3)

    The second form is an example of the real utility of this coercer
    closure: if by error a sequence is passed as it to a function that
    expect a variable number of argument, this coercer fixes it.

    Instance variable `scope` stores the last processed invalid argument.

    When executed, usually `arg` is a tuple received by a function as
    ``*args`` form::

    When executed, returns a tuple, or the same type of source iterable
    argument if possible.

    See `xoutil.params`:mod: for a more specialized and full function
    arguments conformer.

    See `combo`:class: for a combined coercer that  validate each member with
    a separate member coercer.

    '''

    def __init__(self, arg_coerce):
        super(pargs, self).__init__()
        self.inner = check(coercer, arg_coerce)

    def __call__(self, arg):
        from xoutil import Unset
        from xoutil.collections import Iterable
        coerce = self.inner
        if isinstance(arg, Iterable):
            arg = tuple(arg)
            if len(arg) == 1:
                item = arg[0]
                aux = coerce(item)
                if valid(aux):
                    res = (aux, )
                elif isinstance(item, Iterable):
                    res = Unset
                    arg = tuple(item)
                else:
                    self.scope = item
                    res = Invalid
            else:
                res = Unset
            if res is Unset:
                res = arg
                i = 0
                while valid(res) and i < len(res):
                    item = res[i]
                    new = coerce(item)
                    if valid(new):
                        if new is not item:
                            if isinstance(res, tuple):
                                res = list(res)
                            res[i] = new
                        i += 1
                    else:
                        self.scope = item
                        res = Invalid
                if valid(res):
                    res = tuple(res)
        else:
            res = Invalid
        return res


class iterable(custom):
    '''Create a inner coercer that coerces an `iterable` member a member.

    See constructor for more information.

    Return a list, or the same type of source iterable argument if possible.

    For example::

      >>> from xoutil.values import (iterable, int_coerce,
      ...                            create_unique_member_coerce)

      >>> sample = {'1', 1, '1.0'}

      >>> sc = iterable(int_coerce)
      >>> sc(set(sample)) == {1}
      True

    See `mapping`:class: for more details of this problem.  The equivalent
    safe example is::

      >>> member_coerce = create_unique_member_coerce(int_coerce, sample)
      >>> sc = iterable(member_coerce)
      >>> sc(set(sample))
      Invalid

    when executed coerces `arg` (an iterable) member a member using
    `member_coercer`. If any member coercion fails, the full execution also
    fails.

    There are three types of results when an instance is executed:
    (1) iterables that are coerced without modifications, (2) the modified
    ones but conserving its type, and (3) those that are returned in a list.

    '''

    def __init__(self, member_coerce, outer_coerce=None):
        '''Constructor for `iterable`:class: coercers.

        :param member_coerce: A coerce to check each iterable member.

        :param outer_coerce: A coerce to check the type of the entire
               iterable.  Normally a type or tuple of types subclases of
               ``collections.Iterable``.

        '''
        super(iterable, self).__init__()
        member_coerce = check(coercer, member_coerce)
        outer_coerce = compose(coercer(outer_coerce), sized_coerce)
        self.inner = (member_coerce, outer_coerce)

    def __call__(self, arg):
        from xoutil.collections import (Set, Sequence, MutableSequence)
        member_coerce, outer_coerce = self.inner
        modified = False
        aux = outer_coerce(arg)
        if valid(aux):
            arg = aux
            if isinstance(arg, Sequence):
                res = arg
                retyped = False
                mutable = isinstance(arg, MutableSequence)
            else:
                res = list(arg)
                retyped = mutable = True
            i = 0
            while valid(res) and i < len(res):
                item = res[i]
                new = member_coerce(item)
                if valid(new):
                    if new is not item:
                        if not mutable:
                            res = list(res)
                            retyped = mutable = True
                        res[i] = new
                        modified = True
                    i += 1
                else:
                    self.scope = item
                    res = Invalid
            if valid(res):
                if isinstance(arg, Set) and not modified:
                    res = arg
                elif retyped:
                    try:
                        res = type(arg)(res)
                    except BaseException:
                        pass
        else:
            self.scope = arg
            res = Invalid
        return res


class mapping(custom):
    '''Create a coercer to check dictionaries.

    Receives two coercers, one for keys and one for values.

    For example::

      >>> from xoutil.values import (mapping, int_coerce, float_coerce,
      ...                            create_unique_member_coerce)

      >>> sample = {'1': 1, 2.0: '3', 1.0 + 0j: '4.1'}

      >>> dc = mapping(int_coerce, float_coerce)
      >>> dc(dict(sample)) == {1: 1.0, 2: 3.0}
      True

    When coercing containers it's probable that members become repeated after
    coercing them.  This could be not desirable (mainly in sets and
    dictionaries). In those cases use `create_unique_member_coerce`:func: to
    wrap member coercer.  For example::

      >>> key_coerce = create_unique_member_coerce(int_coerce, sample)
      >>> dc = mapping(key_coerce, float_coerce)
      >>> dc(dict(sample))
      Invalid

    Above problem is because it's the same integer (same hash) coerced
    versions of ``'1'`` and ``1.0+0j``.

    This problem of objects of different types that have the same hash is a
    problem to use a example as below::

      >>> {1: int, 1.0: float, 1+0j: complex} == {1: complex}
      True

    '''
    _str_join = _repr_join = ':'

    def __new__(cls, key_coercer=None, value_coercer=None):
        '''Constructor for `mapping`:class: coercers.

        :param key_coercer: A coerce to check each one of the mapping keys.

        :param value_coercer: A coerce to check each one of corresponding
               mapping values.

        '''
        from xoutil.collections import Mapping
        if key_coercer is value_coercer is None:
            return coercer(Mapping)
        else:
            self = super(mapping, cls).__new__(cls)
            key_coercer = check(coercer, key_coercer)
            value_coercer = check(coercer, value_coercer)
            self.inner = (key_coercer, value_coercer)
            return self

    def __call__(self, arg):
        from xoutil.collections import (Mapping, MutableMapping)
        if isinstance(arg, Mapping):
            key_coercer, value_coercer = self.inner
            res = arg
            retyped = False
            mutable = isinstance(arg, MutableMapping)
            keys = list(res)
            i = 0
            while valid(res) and i < len(keys):
                key = keys[i]
                value = res[key]
                new_key = key_coercer(key)
                if valid(new_key):
                    new_value = value_coercer(value)
                    if valid(new_value):
                        if new_key is not key or new_value is not value:
                            if not mutable:
                                res = dict(res)
                                retyped = mutable = True
                            if key is not new_key:
                                del res[key]
                            res[new_key] = new_value
                        i += 1
                    else:
                        self.scope = ({key: value}, value_coercer)
                        res = Invalid
                else:
                    self.scope = ({key: value}, key_coercer)
                    res = Invalid
            if valid(res) and retyped:
                try:
                    res = type(arg)(res)
                except BaseException:
                    pass
        else:
            self.scope = ()
            res = Invalid
        return res


del sys, ABCMeta, metaclass, regex_compile, lwraps



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.eight.exceptions
# ---------------------------------------------------------------------
# Copyright (c) 2015 Merchise and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2015-10-14

'''Compatibility exceptions between Python 2 and 3.

Python 2 defines an module named `exceptions` but Python 3 doesn't.  Also,
there are some exception classes defined in Python 2 but not in Python 3.


'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


from . import _py3, exec_


try:
    from exceptions import StandardError
except ImportError:
    StandardError = Exception

try:
    BaseException = BaseException
except NameError:
    BaseException = StandardError


if _py3:
    def reraise(tp, value, tb=None):
        if value.__traceback__ is not tb:
            raise value.with_traceback(tb)
        raise value
else:
    exec_("""def reraise(tp, value, tb=None): raise tp, value, tb""")

reraise.__doc__ = '''Raise an exception with a traceback.

In Python 2.7 this is the same as::

    raise T, V, tb

In Python 3+ this is::

    raise T(V).with_traceback(tb)

.. versionadded: 1.7.1

'''

del _py3, exec_



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.ids
# ---------------------------------------------------------------------
# Copyright (c) 2015-2017 Merchise and Contributors
# Copyright (c) 2013, 2014 Merchise Autrement and Contributors
# Copyright (c) 2012 Medardo Rodríguez
# All rights reserved.
#
# Contributors: see CONTRIBUTORS and HISTORY file
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2015-08-25

'''Utilities to obtain identifiers that are unique at different contexts.

Contexts could be global, host local or application local.  All standard
`uuid`:mod: tools are included in this one: `UUID`:class:, `uuid1`:func:,
`uuid3`:func:, `uuid4`:func:, `uuid5`:func:, `getnode`:func:` and standard
UUIDs constants `NAMESPACE_DNS`, `NAMESPACE_URL`, `NAMESPACE_OID` and
`NAMESPACE_X500`.

This module also contains:

- `str_uuid`:func:\ : Return a string with a GUID representation, random if
  the argument is True, or a host ID if not.

- `slugify`:func:\ : Convert any object to a valid slug (mainly used with
  string parameters).

.. versionadded:: 1.7.0

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_imports)


from uuid import (UUID, uuid1, uuid3, uuid4, uuid5, getnode,    # noqa
                  NAMESPACE_DNS, NAMESPACE_URL, NAMESPACE_OID, NAMESPACE_X500)


def str_uuid(random=False):
    '''Return a "Global Unique ID" as a string.

    :param random: If True, a random uuid is generated (does not use host id).

    '''
    fn = uuid4 if random else uuid1
    return '{}'.format(fn())


def slugify(value, *args, **kwargs):
    '''Return the string normal form, valid for slugs, for a given value.

    All values expecting strings are coerced using the following methodology:
    normalize all non-ascii to valid characters using unicode 'NFKC', converts
    to lower-case and translate all invalids (including white-spaces) to the
    character defined as `replacement` (normally a hyphen, see below).

    In all values expecting sets repeated characters will be removed.

    In the result value, all repeated replacement occurrences will be striped;
    as well as those in the leading and trailing positions.

    :param value: The source value to slugify.  It's the first positional
           argument and always is required.

    :param replacement: A character to be used as replacement for unwanted
           characters.

           Could be given as a name argument or in the second positional
           argument (index 0 in `args`).  Default value is a hyphen ('-').

    :param invalids: Characters to be considered invalids.  There is a default
           set of valid characters which are kept in the resulting slug.
           Characters given in this parameter are removed from the resulting
           valid character set (see next parameter).

           Extra argument values can be used for compatibility with
           `invalid_underscore` argument in deprecated `normalize_slug`:func:
           xoutil function:

           - ``True`` is a synonymous of underscore ``"_"``.

           - ``False`` or ``None``: An empty set.

           Could be given as a name argument or in the third positional
           argument (index 1 in `args`).  Default value is an empty set.

    :param valids: A collection of extra valid characters.  Default set of
           valid characters are alpha-numeric and the underscore (those that
           fulfill with the regular expression ``[_a-z0-9]``).  All characters
           not convertible to ascii are ignored.

           This argument is processed before `invalids`.

           Could be given only as a name argument.  Default value is an empty
           set.

    :param smalls: Management for small parts, those with 3 or less
           characters.  Next, all possible values for this argument:

           1. ``None``, ``False`` or the string ``"crop"``: remove small
              parts.

           2. ``True`` or the string ``"keep"``: maintain small parts; this is
              the default value.

           3. The string ``"join"``: small parts are joined to the smallest
              of its consecutive parts.

    Parameters `value` and `replacement` could be of any (non-string) type,
    these values are normalized and converted to lower-case ASCII strings.

    Examples::

      >>> slugify('  Á.e i  Ó  u  ') == 'a-e-i-o-u'
      True

      >>> slugify('  Á.e i  Ó  u  ', '.', invalids='AU') == 'e.i.o'
      True

      >>> slugify('  Á.e i  Ó  u  ', valids='.') == 'a.e-i-o-u'
      True

      >>> slugify('_x', '_') == '_x'
      True

      >>> slugify('-x', '_') == 'x'
      True

      >>> slugify(None) == 'none'
      True

      >>> slugify(1 == 1)  == 'true'
      True

      >>> slugify(1.0) == '1-0'
      True

      >>> slugify(135) == '135'
      True

      >>> slugify(123456, replacement='', invalids='52') == '1346'
      True

      >>> slugify('_x', replacement='_') == '_x'
      True

    .. versionadded:: 1.7.0 To deprecate `normalize_slug` and add
                      `crop_smalls` parameter.

    '''
    import re
    from xoutil.eight import string_types
    from xoutil.string import normalize_ascii as ascii
    # local functions
    lascii = lambda v: ascii(v).lower()
    _set = lambda v: ''.join(set(v))
    _esc = lambda v: re.escape(_set(v))
    _from_iter = lambda v: ''.join(i for i in v)
    # check and adjust arguments
    if replacement in (None, False):
        replacement = ''
    elif isinstance(replacement, string_types):
        replacement = ascii(replacement)    # TODO: or lascii?
    else:
        msg = '`replacement` (%s) must be a string or None, not `%s`.'
        raise TypeError(msg % (replacement, type(replacement)))
    if invalids is True:
        # Backward compatibility with former `invalid_underscore` argument
        invalids = '_'
    elif invalids in {None, False}:
        invalids = ''
    else:
        if not isinstance(invalids, string_types):
            invalids = _from_iter(invalids)
        invalids = _esc(lascii(invalids))
    if valids is None:
        valids = ''
    else:
        if not isinstance(valids, string_types):
            valids = _from_iter(valids)
        valids = _esc(re.sub(r'[0-9a-b]+', '', lascii(valids)))
    # calculate result
    res = lascii(value)
    regex = re.compile(r'[^_a-z0-9%s]+' % valids)
    repl = '\t' if replacement else ''
    res = regex.sub(repl, res)
    if invalids:
        regex = re.compile(r'[%s]+' % invalids)
        res = regex.sub(repl, res)
    if repl:
        r = {'r': r'%s' % re.escape(repl)}
        regex = re.compile(r'(%(r)s){2,}' % r)
        res = regex.sub(repl, res)
        regex = re.compile(r'(^%(r)s+|%(r)s+$)' % r)
        res = regex.sub('', res)
        regex = re.compile(r'[\t]' % r)
        res = regex.sub(replacement, res)
    return res



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.logical
# ---------------------------------------------------------------------
# Copyright (c) 2015 Merchise and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under
# the terms of the LICENCE attached in the distribution package.
#
# Created on 2015-02-23

'''Fixed special logical values like Unset, Ignored, Required, etc...

All values only could be `True` or `False` but are intended in places where
`None` is expected to be a valid value or for special Boolean formats.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        unicode_literals as _py3_unicode,
                        absolute_import as _py3_abs_imports)


# TODO: Maybe `UnsetType` can be deprecated after this module.

from abc import ABCMeta
from .eight.meta import metaclass


class MetaLogical(ABCMeta):
    '''Metaclass for logical types.'''

    def __init__(self, name, bases, attrs):
        # from weakref import WeakValueDictionary as maptype
        maptype = dict    # TODO: weak refs not working with slots in Python3
        if issubclass(self, int):    # ``bool`` is not allowed as a base class
            super(MetaLogical, self).__init__(name, bases, attrs)
            self.__instances__ = maptype()
            self.register(bool)
        else:
            raise TypeError('Only `int` sub-classes are allowed!')

    def _get_instance(self, name, value):
        '''Get or create a new logical instance.

        :param name: String representing the internal name.  Logical instances
               are unique (singletons) in the context of this argument.  ``#``
               and spaces are invalid characters to allow comments.

        :param value: Any value compatible with Python `bool` (Always is
               converted to this type before using it).  Default is `False`.

        This method is only intended to be called from `__new__` special
        method implementations.

        '''
        from .eight import intern as unique
        name = unique(name)
        if name:
            value = bool(value)
            res = self.__instances__.get(name)
            if res is None:    # Create the new instance
                res = super(self, self).__new__(self, value)
                self.__instances__[name] = res
            elif res != value:    # Check existing instance
                msg = 'Incorrect value for existing instance: %s '
                raise ValueError(msg % name)
            return res
        else:
            raise ValueError('Name must be a valid non empty string!')

    def get_name(self, value):
        '''Returns name of a given logical instance (:param value:).

        Standard Python Boolean values are valid too.

        '''
        try:
            d = self.__instances__
            items = ((name, d[name]) for name in d)
            return next(name for name, obj in items if obj is value)
        except StopIteration:
            return repr(bool(value))

    def parse(self, value):
        '''Returns instance (or None if not found) from a string.

        Standard Python Boolean values are parsed too.

        '''
        if '#' in value:    # Remove comment
            value = value.split('#')[0].strip()
        key = next((key for key in self.__instances__ if key == value), None)
        if key is not None:
            return self.__instances__[key]
        else:
            try:    # Standard Python Boolean values
                res = eval(value)
                if not (res is True or res is False):
                    res = None
            except:
                res = None
            return res


class Logical(int, metaclass(MetaLogical)):
    '''Instances are custom logical values (`True` or `False`).

    See :meth:`~MetaLogical._get_instance` method for information on
    constructor arguments.

    For example::

      >>> true = Logical('true', True)
      >>> false = Logical('false')
      >>> none = Logical('false')
      >>> unset = Logical('unset')

      >>> class X(object):
      ...      attr = None

      >>> getattr(X(), 'attr') is not None
      False

      >>> getattr(X(), 'attr', false) is not false
      True

      >>> none is false
      True

      >>> false == False
      True

      >>> false == unset
      True

      >>> false is unset
      False

      >>> true == True
      True

    '''

    __slots__ = ()

    def __new__(cls, name, value=False):
        '''Constructor of a new instance.'''
        return cls._get_instance(name, value)

    def __repr__(self):
        return type(self).get_name(self)

    __str__ = __repr__


del ABCMeta, metaclass



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.string
# ---------------------------------------------------------------------
# Copyright (c) 2015-2017 Merchise and Contributors
# Copyright (c) 2013, 2014 Merchise Autrement and Contributors
# Copyright (c) 2012 Medardo Rodriguez
# All rights reserved.
#
# Author: Medardo Rodríguez
# Contributors: see CONTRIBUTORS and HISTORY file
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on Feb 17, 2012

'''Exposes all original `string` module functionalities, with some general
additions.

In this module `str` and `unicode` types are not used because Python 2.x and
Python 3.x treats strings differently.  `bytes` and `text_type` will be used
instead with the following conventions:

- In Python 2.x `str` is synonym of `bytes` and both (`unicode` and 'str') are
  both string types inheriting form `basestring`.

- In Python 3.x `str` is always unicode but `unicode` and `basestring` types
  doesn't exists. `bytes` type can be used as an array of one byte each item.

  Many methods are readjusted to these conditions.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        # unicode_literals as _py3_unicode,
                        absolute_import as _py3_abs_imports)

from xoutil.deprecation import deprecated as _deprecated
from xoutil.eight import _py3

from xoutil.modules import copy_members as _copy_python_module_members
_pm = _copy_python_module_members()

Formatter = _pm.Formatter     # Redundant but needed to avoid IDE errors

del _copy_python_module_members, _pm


def force_encoding(encoding=None):
    '''Validates an encoding value; if None use `locale.getlocale()[1]`; else
    return the same value.

    .. versionadded:: 1.2.0

    '''
    # TODO: Maybe use only `sys.getdefaultencoding()`
    import locale
    return encoding or locale.getpreferredencoding() or 'UTF-8'


def safe_decode(s, encoding=None):
    '''Similar to bytes `decode` method returning unicode.

    Decodes `s` using the given `encoding`, or determining one from the system.

    Returning type depend on python version; if 2.x is `unicode` if 3.x `str`.

    .. versionadded:: 1.1.3

    '''
    from xoutil.eight import text_type
    if isinstance(s, text_type):
        return s
    else:
        encoding = force_encoding(encoding)
        try:
            # In Python 3 str(b'm') returns the string "b'm'" and not just "m",
            # this fixes this.
            return text_type(s, encoding, 'replace')
        except LookupError:
            # The provided enconding is not know, try with no encoding.
            return safe_decode(s)
        except:
            # For numbers and other stuff.
            return text_type(s)


def safe_encode(u, encoding=None):
    '''Similar to unicode `encode` method returning bytes.

    Encodes `u` using the given `encoding`, or determining one from the system.

    Returning type is always `bytes`; but in python 2.x is also `str`.

    .. versionadded:: 1.1.3

    '''
    # TODO: This is not nice for Python 3, bytes is not valid string any more
    #       See :func:`json.encoder.py_encode_basestring_ascii` of Python 2.x
    from xoutil.eight import string_types, text_type
    if isinstance(u, bytes):
        return u
    else:
        encoding = force_encoding(encoding)
        try:
            try:
                if isinstance(u, string_types):
                    # In Python 2.x bytes does not allows an encoding argument.
                    return bytes(u)
                else:
                    return text_type(u).encode(encoding, 'replace')
            except:
                return text_type(u).encode(encoding, 'replace')
        except LookupError:
            return safe_encode(u)


def safe_str(obj=str()):
    '''Convert to normal string type in a safe way.

    Most of our Python 2.x code uses unicode as normal string, also in
    Python 3 converting bytes or byte-arrays to strings includes the "b"
    prefix in the resulting value.

    This function is useful in some scenarios that require `str` type (for
    example attribute ``__name__`` in functions and types).

    As ``str is bytes`` in Python2, using str(value) assures correct these
    scenarios in most cases, but in other is not enough, for example::

      >>> from xoutil.string import safe_str as sstr
      >>> def inverted_partial(func, *args, **keywords):
      ...     def inner(*a, **kw):
      ...         a += args
      ...         kw.update(keywords)
      ...         return func(*a, **kw)
      ...     inner.__name__ = sstr(func.__name__.replace('lambda', u'λ'))
      ...     return inner

    .. versionadded:: 1.7.0

    '''
    if _py3:
        if isinstance(obj, (bytes, bytearray)):
            return safe_decode(obj)
        else:
            return str(obj)
    else:
        try:
            return str(obj)
        except UnicodeEncodeError:
            # assert isinstance(value, unicode)
            return safe_encode(obj)


def safe_join(separator, iterable, encoding=None):
    '''Similar to `join` method in string objects `separator.join(iterable)`, a
    string which is the concatenation of the strings in the `iterable` with
    `separator` as intermediate between elements. Return unicode or bytes
    depending on type of `separator` and each item in `iterable`.

    `encoding` is used in case of error to concatenate bytes + unicode.

    This function must be deprecated in Python 3.

    .. versionadded:: 1.1.3

    .. warning:: The `force_separator_type` was removed in version 1.2.0.

    '''
    try:
        return separator.join(iterable)
    except:
        pass
    encoding = force_encoding(encoding)
    empty = True
    for item in iterable:
        if empty:
            res = item
            empty = False
        else:
            for tail in (separator, item):
                try:
                    res += tail
                except:
                    res = (safe_decode(res, encoding) +
                           safe_decode(item, encoding))
    return res if not empty else type(separator)()


# Makes explicit the deprecation warning for py3k.
if _py3:
    safe_join = _deprecated('builtin join method of str',
                            'safe_join is deprecated for Python 3. Use '
                            'builtin join method of str.')(safe_join)


def safe_strip(value):
    '''Removes the leading and tailing space-chars from `value` if string, else
    return `value` unchanged.

    .. versionadded:: 1.1.3

    '''
    from xoutil.eight import string_types
    return value.strip() if isinstance(value, string_types) else value


def cut_prefix(value, prefix):
    '''Removes the leading `prefix` if exists, else return `value`
    unchanged.

    '''
    from xoutil.eight import text_type as str, binary_type as bytes
    if isinstance(value, str) and isinstance(prefix, bytes):
        prefix = safe_decode(prefix)
    elif isinstance(value, bytes) and isinstance(prefix, str):
        prefix = safe_encode(prefix)
    return value[len(prefix):] if value.startswith(prefix) else value


def cut_any_prefix(value, *prefixes):
    '''Apply `cut_prefix`:func: for the first matching prefix.'''
    result = prev = value
    i, top = 0, len(prefixes)
    while i < top and result == prev:
        prefix, i = prefixes[i], i + 1
        prev, result = result, cut_prefix(prev, prefix)
    return result


def cut_prefixes(value, *prefixes):
    '''Apply `cut_prefix`:func: for all provided prefixes in order.'''
    result = value
    for prefix in prefixes:
        result = cut_prefix(result, prefix)
    return result


def cut_suffix(value, suffix):
    '''Removes the tailing `suffix` if exists, else return `value`
    unchanged.

    '''
    from xoutil.eight import text_type as str, binary_type as bytes
    if isinstance(value, str) and isinstance(suffix, bytes):
        suffix = safe_decode(suffix)
    elif isinstance(value, bytes) and isinstance(suffix, str):
        suffix = safe_encode(suffix)
    # Since value.endswith('') is always true but value[:-0] is actually
    # always value[:0], which is always '', we have to explictly test for
    # len(suffix)
    if len(suffix) > 0 and value.endswith(suffix):
        return value[:-len(suffix)]
    else:
        return value


def cut_any_suffix(value, *suffixes):
    '''Apply `cut_suffix`:func: for the first matching suffix.'''
    result = prev = value
    i, top = 0, len(suffixes)
    while i < top and result == prev:
        suffix, i = suffixes[i], i + 1
        prev, result = result, cut_suffix(prev, suffix)
    return result


def cut_suffixes(value, *suffixes):
    '''Apply `cut_suffix`:func: for all provided suffixes in order.'''
    result = value
    for suffix in suffixes:
        result = cut_suffix(result, suffix)
    return result


def capitalize_word(value):
    'Capitalizes the first char of value'
    if value and value[0].islower():
        return value[0].upper() + value[1:]
    else:
        return value


def capitalize(value, title=True):
    '''Capitalizes value according to whether it should be title-like.

    Title-like means it will capitalize every word but the 3-letters or less
    unless its the first word::

        >>> capitalize('a group is its own worst enemy')
        'A Group is its own Worst Enemy'

    (This may be odd because, in the example above, own should be capitalized.)

    Return bytes or unicode depending on type of `value`.

        >>> from xoutil.eight import text_type
        >>> type(capitalize(text_type('something'))) is text_type
        True

        >>> type(capitalize(str('something'))) is str
        True

    '''
    tstr = type(value)
    space, empty = tstr(' '), tstr('')
    words = value.split() if value else None
    if words:
        count = len(words) if title else 1
        for i in range(count):
            word = words[i]
            if len(word) > 3 or i == 0:
                word = capitalize_word(word)
                words[i] = word
        return space.join(words)
    else:
        return empty


def hyphen_name(name):
    '''Convert a name, normally an identifier, to a hyphened slug.

    All transitions from lower to upper capitals (or from digits to letters)
    are joined with a hyphen.

    Also, all invalid characters (those invalid in Python identifiers) are
    converted to hyphens.

    For example::

      >>> hyphen_name('BaseNode') == 'base-node'
      True

    '''
    import re
    regex = re.compile('([a-z0-9][A-Z]|[a-zA-Z][0-9]|[0-9][a-z])')
    parts = []
    for m in reversed(list(regex.finditer(name))):
        i, f = m.span()
        name, tail = name[:i + 1], name[i + 1:]
        parts.insert(0, tail)
    parts.insert(0, name)
    name = '-'.join(parts)
    return safe_str(normalize_slug(name, '-', '_'))


# TODO: Document and fix all these "normalize_..." functions
def normalize_unicode(value):
    # FIXME: i18n
    if (value is None) or (value is str('')):
        return ''
    elif value is True:
        return safe_decode('Sí')
    elif value is False:
        return safe_decode('No')
    else:
        return safe_decode(value)


def normalize_name(value):
    return capitalize(normalize_unicode(value))


def normalize_title(value):
    return capitalize(normalize_unicode(value), True)


def normalize_str(value):
    import re
    is_bytes = isinstance(value, bytes)
    regex, sep = r'(\S+)\s*', ' '
    if is_bytes:
        regex, sep = bytes(regex), bytes(sep)
    regex = re.compile(regex)
    matches = regex.findall(value)
    names = (m.capitalize() if len(m) >= 3 else m.lower() for m in matches)
    return sep.join(names)


def normalize_ascii(value):
    '''Return the string normal form for the `value`

    Convert all non-ascii to valid characters using unicode 'NFKC'
    normalization.

    '''
    import unicodedata
    from xoutil.eight import text_type
    if not isinstance(value, text_type):
        value = safe_decode(value)
    res = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore')
    return safe_str(res)


def normalize_slug(value, replacement='-', invalids=None, valids=None):
    '''Return the string normal form, valid for slugs, for the `value`

    Convert all non-ascii to valid characters using unicode 'NFKC'
    normalization.

    Lower-case the result.

    Replace unwanted characters by `replacement`, repetition of given pattern
    will be converted to only one instance.

    .. warning:: There's a known bug when `replacement` contains '\'.

    ``[_a-z0-9]`` are assumed as valid characters.  Extra arguments can modify
    this standard behaviour:

    :param invalids: Any collection of characters added to these that are
           normally invalid in the provided `value`. (non-ascii or not
           included in valid characters).  Boolean ``True`` can be passed as a
           synonymous of ``"_"`` for compatibility with old
           ``invalid_underscore`` argument.  ``False`` or ``None`` are assumed
           as an empty set for invalid characters.

    :param valids: A collection of extra valid characters (all non-ascii
           characters are ignored).  This parameter could be either a valid
           string, any iterator of valid strings of characters, or ``None`` to
           use only default valid characters (See above).

    .. warning:: The result may contain characters in `invalids` if
                 `replacements` does.

    Parameters `value` and `replacement` could be of any (non-string) type,
    these values are normalized and converted to lower-case ASCII strings.

    Examples::

      >>> normalize_slug('  Á.e i  Ó  u  ') == 'a-e-i-o-u'
      True

      >>> normalize_slug('  Á.e i  Ó  u  ', '.', invalids='AU') == 'e.i.o'
      True

      >>> normalize_slug('  Á.e i  Ó  u  ', valids='.') == 'a.e-i-o-u'
      True

      >>> normalize_slug('_x', '_') == '_x'
      True

      >>> normalize_slug('-x', '_') == 'x'
      True

      >>> normalize_slug(None) == 'none'
      True

      >>> normalize_slug(1 == 1)  == 'true'
      True

      >>> normalize_slug(1.0) == '1-0'
      True

      >>> normalize_slug(135) == '135'
      True

      >>> normalize_slug(123456, '', invalids='52') == '1346'
      True

      >>> normalize_slug('_x', '_') == '_x'
      True

    .. versionchanged:: 1.5.5 Added the `invalid_underscore` parameter.

    .. versionchanged:: 1.6.6 Replaced the `invalid_underscore` paremeter by
       `invalids`.  Added the `valids` parameter.

    .. versionchanged:: 1.7.2 Clarified the role of `invalids` with regards to
       `replacement`.

    '''
    import re
    from xoutil.eight import string_types
    # local functions
    _normalize = lambda v: normalize_ascii(v).lower()
    _set = lambda v: ''.join(set(v))
    _esc = lambda v: re.escape(_set(v))
    _from_iter = lambda v: ''.join(i for i in v)
    # check and adjust arguments
    if replacement in (None, False):
        replacement = ''
    elif isinstance(replacement, string_types):
        replacement = normalize_ascii(replacement)    # TODO: or _normalize?
    else:
        msg = '`replacement` (%s) must be a string or None, not `%s`.'
        raise TypeError(msg % (replacement, type(replacement)))
    if invalids is True:
        # Backward compatibility with former `invalid_underscore` argument
        invalids = '_'
    elif invalids in {None, False}:
        invalids = ''
    else:
        if not isinstance(invalids, string_types):
            invalids = _from_iter(invalids)
        invalids = _esc(_normalize(invalids))
    if valids is None:
        valids = ''
    else:
        if not isinstance(valids, string_types):
            valids = _from_iter(valids)
        valids = _esc(re.sub(r'[0-9a-b]+', '', _normalize(valids)))
    # calculate result
    res = _normalize(value)
    regex = re.compile(r'[^_a-z0-9%s]+' % valids)
    repl = '\t' if replacement else ''
    res = regex.sub(repl, res)
    if invalids:
        regex = re.compile(r'[%s]+' % invalids)
        res = regex.sub(repl, res)
    if repl:
        r = {'r': r'%s' % re.escape(repl)}
        regex = re.compile(r'(%(r)s){2,}' % r)
        res = regex.sub(repl, res)
        regex = re.compile(r'(^%(r)s+|%(r)s+$)' % r)
        res = regex.sub('', res)
        regex = re.compile(r'[\t]' % r)
        res = regex.sub(replacement, res)
    return res


def strfnumber(number, format_spec='%0.2f'):
    res = format_spec % number
    if '.' in res:
        res = res.rstrip('0')
        if res.endswith('.'):
            res = res[:-1]
    return res


def parse_boolean(value):
    '''Parse a boolean from any value given a special treatment to
    strings.

    >>> parse_boolean('trUe')
    True

    >>> parse_boolean('faLSe')
    False

    '''
    from xoutil.eight import string_types
    if isinstance(value, string_types):
        value = value.strip()
        if value:
            if value.isdigit():
                return bool(int(value))
            else:
                if isinstance(value, bytes):
                    falses = (b'false', b'no', b'not')
                else:
                    falses = ('false', 'no', 'not')
                return value.lower() not in falses
        else:
            return False
    else:
        return bool(value)


def parse_url_int(value, default=None):
    '''Parse an integer URL argument. Some operations treat simple
    arguments as a list of one element.

    '''
    # TODO: Move to `xoutil.web`
    if isinstance(value, (list, tuple, set)) and len(value) > 0:
        value = value[0]
    try:
        return int(safe_strip(value))
    except:
        return default


def error2str(error):
    '''Convert an error to string.'''
    from xoutil.eight import string_types
    from xoutil.types import type_coerce
    if isinstance(error, string_types):
        return safe_str(error)
    elif isinstance(error, BaseException):
        tname = type(error).__name__
        res = safe_str(error)
        if tname in res:
            return res
        else:
            return str(': ').join(tname, res) if res else tname
    elif issubclass(error, BaseException):
        return type(error).__name__
    else:
        prefix = str('unknown error: ')
        cls = type_coerce(error)
        tname = cls.__name__
        if cls is error:
            res = tname
        else:
            res = safe_str(error)
            if tname not in res:
                res = str('{}({})').format(tname, res) if res else tname
        return prefix + res


def force_str(value, encoding=None):
    '''Force to string, the type is different in Python 2 or 3 (bytes or
    unicode).

    :param value: The value to convert to `str`.
    :param encoding: The encoding which should be used if either encoding
                     or decoding should be performed on `value`.

                     The default is to use the same default as
                     :func:`safe_encode` or :func:`safe_decode`.

    .. versionadded:: 1.2.0

    '''
    if isinstance(value, str):
        return value
    elif str is bytes:      # Python 2
        return safe_encode(value, encoding)
    else:
        return safe_decode(value, encoding)


def make_a10z(string):
    '''Utility to find out that "internationalization" is "i18n".

    Examples::

       >>> print(make_a10z('parametrization'))
       p13n
    '''
    return string[0] + str(len(string[1:-1])) + string[-1]


from xoutil.eight import input

input = _deprecated(
    input,
    "xoutil.string.input is deprecated.  Use xoutil.eight.input"
)(input)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.tools
# ---------------------------------------------------------------------
# Copyright (c) 2015 Merchise and Contributors
# Copyright (c) 2014 Merchise Autrement and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under
# the terms of the LICENCE attached in the distribution package.
#
# Created 2014-05-15

'''Simple tools with no dependencies on other modules.

Extensions to the Python's standard library.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        unicode_literals as _py3_unicode,
                        absolute_import as _py3_abs_imports)


def get_default(args, default=None):
    '''Get a default value passed as last positional argument.

    Several functions that get values define an optional default value
    parameter.  To use a construction ``def get_foobar(name, default=None)``
    sometimes is not possible because `None` could be a possible valid
    "foobar" value.  In these cases it's better to construct something like::

      def get_foobar(name, *default):
          ...

    and in client code you can call this function with a impossible "foobar"
    value, like: ``res = get_foobar('egg', Undefined)`` (see `xoutil.values`).

    This function receive the tuple as received by the function, and which
    value to return if None is given::

      def get_foobar(name, *default):
          from xoutil.tools import get_default
          from xoutil import Undefined as _undef
          default = get_default(default, _undef)
          ...

    If tuple `args` has an invalid size, a `TypeError` exception is raised.

    '''
    count = len(args)
    if count == 0:
        return default
    elif count == 1:
        return args[0]
    else:
        raise TypeError("expected 0 or 1 values for default, got %s" % count)



# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.decorator
# ---------------------------------------------------------------------
# Copyright (c) 2015 Merchise and Contributors
# Copyright (c) 2013, 2014 Merchise Autrement and Contributors
# Copyright (c) 2009-2012 Medardo Rodríguez
#
# Author: Medardo Rodriguez
# Contributors: see CONTRIBUTORS and HISTORY file
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#

'''Some useful decorators.'''

# TODO: reconsider all this module


from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        unicode_literals as _py3_unicode,
                        absolute_import)

import sys

from functools import wraps
from types import FunctionType as function

from .meta import decorator as _decorator


from xoutil.names import strlist as strs
__all__ = strs('decorator', 'AttributeAlias', 'settle', 'namer', 'aliases',
               'assignment_operator', 'instantiate', 'memoized_property',
               'memoized_instancemethod', 'reset_memoized')
del strs


class AttributeAlias(object):
    '''Descriptor to create aliases for object attributes.

    This descriptor is mainly to be used internally by :func:`aliases`
    decorator.

    '''

    def __init__(self, attr_name):
        super(AttributeAlias, self).__init__()
        self.attr_name = attr_name

    def __get__(self, instance, owner):
        return getattr(instance or owner, self.attr_name)

    def __set__(self, instance, value):
        setattr(instance, self.attr_name, value)

    def __delete__(self, instance):
        delattr(instance, self.attr_name)


def settle(**kwargs):
    '''Returns a decorator that sets different attribute values to the
    decorated target (function or class).

    Usage::

       >>> @settle(name='Name')
       ... class Person(object):
       ...    pass

       >>> Person.name
       'Name'

    '''
    def inner(target):
        from xoutil.eight import iteritems
        for key, value in iteritems(kwargs):
            setattr(target, key, value)
        return target
    return inner


def namer(name, **kwargs):
    '''Like :func:`settle`, but name is a positional argument and is assigned
    to the attribute ``__name__``.

    Usage::

        >>> @namer('Identity', custom=1)
        ... class I(object):
        ...    pass

        >>> I.__name__
        'Identity'

        >>> I.custom
        1

    '''
    return settle(__name__=name, **kwargs)


def aliases(*names, **kwargs):
    '''In a class, create an :class:`AttributeAlias` descriptor for each
    definition as keyword argument (alias=existing_attribute).

    If "names" are given, then the definition context is looked and are
    assigned to it the same decorator target with all new names::

        >>> @aliases('foo', 'bar')
        ... def foobar(*args):
        ...     'This function is added to its module with two new names.'

    '''
    def inner(target):
        '''Direct closure decorator that settle several attribute aliases.
        '''
        if kwargs:
            assert isinstance(target, type), '"target" must be a class.'
        if names:
            _locals = sys._getframe(1).f_locals
            for name in names:
                _locals[str(name)] = target
        if kwargs:
            for alias, field in kwargs.iteritems():
                setattr(target, alias, AttributeAlias(field))
        return target
    return inner


@_decorator
def assignment_operator(func, maybe_inline=False):
    '''Makes a function that receives a name, and other args to get its first
    argument (the name) from an assigment operation, meaning that it if its
    used in a single assignment statement the name will be taken from the left
    part of the ``=`` operator.

    .. warning:: This function is dependant of CPython's implementation of the
                 language and won't probably work on other implementations.
                 Use only you don't care about portability, but use sparingly
                 (in case you change your mind about portability).

    '''
    import inspect
    import ast

    if not isinstance(func, function):
        raise TypeError('"func" must be a function.')

    @wraps(func)
    def inner(*args):
        frm = sys._getframe(1)
        filename, line, funcname, src_lines, idx = inspect.getframeinfo(frm)
        try:
            sourceline = src_lines[idx].strip()
            parsed_line = ast.parse(sourceline, filename).body[0]
            assert maybe_inline or isinstance(parsed_line, ast.Assign)
            if isinstance(parsed_line, ast.Assign):
                assert len(parsed_line.targets) == 1
                assert isinstance(parsed_line.targets[0], ast.Name)
                name = parsed_line.targets[0].id
            elif maybe_inline:
                assert isinstance(parsed_line, ast.Expr)
                name = None
            else:
                assert False
            return func(name, *args)
        except (AssertionError, SyntaxError):
            if maybe_inline:
                return func(None, *args)
            else:
                return func(*args)
        finally:
            del filename, line, funcname, src_lines, idx
    return inner


@_decorator
def instantiate(target, *args, **kwargs):
    '''Some singleton classes must be instantiated as part of its declaration
    because they represents singleton objects.

    Every argument, positional or keyword, is passed as such when invoking the
    target. The following two code samples show two cases::

       >>> @instantiate
       ... class Foobar(object):
       ...    def __init__(self):
       ...        print('Init...')
       Init...


       >>> @instantiate('test', context={'x': 1})
       ... class Foobar(object):
       ...    def __init__(self, name, context):
       ...        print('Initializing a Foobar instance with name={name!r} '
       ...              'and context={context!r}'.format(**locals()))
       Initializing a Foobar instance with name='test' and context={'x': 1}

    In all cases, Foobar remains the class, not the instance::

        >>> Foobar  # doctest: +ELLIPSIS
        <class '...Foobar'>
    '''
    target(*args, **kwargs)
    return target


del _decorator


# The following is extracted from the SQLAlchemy project's codebase, merit and
# copyright goes to SQLAlchemy authors::
#
# Copyright (C) 2005-2011 the SQLAlchemy authors and contributors
#
# This module is part of SQLAlchemy and is released under the MIT License:
# http://www.opensource.org/licenses/mit-license.php
#
class memoized_property(object):
    """A read-only property that is only evaluated once.

    This is extracted from the SQLAlchemy project's codebase, merit and
    copyright goes to SQLAlchemy authors::

      Copyright (C) 2005-2011 the SQLAlchemy authors and contributors

      This module is part of SQLAlchemy and is released under the MIT License:
      http://www.opensource.org/licenses/mit-license.php

    """
    def __init__(self, fget, doc=None):
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return self
        obj.__dict__[self.__name__] = result = self.fget(obj)
        return result


class memoized_instancemethod(object):
    """Decorate a method memoize its return value.

    Best applied to no-arg methods: memoization is not sensitive to
    argument values, and will always return the same value even when
    called with different arguments.

    This is extracted from the SQLAlchemy project's codebase, merit and
    copyright goes to SQLAlchemy authors::

      Copyright (C) 2005-2011 the SQLAlchemy authors and contributors

      This module is part of SQLAlchemy and is released under the MIT License:
      http://www.opensource.org/licenses/mit-license.php

    """
    def __init__(self, fget, doc=None):
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return self

        def oneshot(*args, **kw):
            result = self.fget(obj, *args, **kw)
            memo = lambda *a, **kw: result
            memo.__name__ = self.__name__
            memo.__doc__ = self.__doc__
            obj.__dict__[self.__name__] = memo
            return result

        oneshot.__name__ = self.__name__
        oneshot.__doc__ = self.__doc__
        return oneshot


def reset_memoized(instance, name):
    instance.__dict__.pop(name, None)


if __name__ == '__main__':
    import doctest
    doctest.testmod(verbose=True)
